import numpy as np
import torch
from typing import Optional
# from scipy.spatial.distance import cdist
from pyannote.core.utils.distance import cdist # This offers more metric options than scipy's cdist

class IncrementalSpeakerClustering:
    def __init__(
        self,
        delta_new: float,
        metric: Optional[str] = "cosine",
        max_speakers: int = 20
    ):
        self.delta_new = delta_new # Threshold to determine if a new centroid is needed
        self.metric = metric # Distance metric, default is "cosine"
        self.max_speakers = max_speakers # Maximum number of global speakers
        self.centers: Optional[np.ndarray] = None # Array to store all centroids
        self.counts = np.zeros(self.max_speakers, dtype=int) # Counter for each centroid

    def get_next_center_position(self) -> Optional[int]:
        """Find the next available position to store a new centroid.

        Returns
        -------
        next_position: int or None
            The index of the next available position; None if no position is available.
        """
        # Find the first vector that is all zeros (i.e., not yet used)
        for i in range(self.max_speakers):
            if np.all(self.centers[i] == 0):
                return i
        return None
    
    def init_centers(self, dimension: int):
        """Initialize the centroid matrix"""
        self.centers = np.zeros((self.max_speakers, dimension))

    def update(self, global_speaker: int, embedding: np.ndarray):
        """Update an existing centroid"""
        assert global_speaker < len(self.centers), "Cannot update unknown centers"
        self.centers[global_speaker] += embedding

    def add_center(self, embedding: np.ndarray) -> int:
        """Add a new centroid and return its index"""
        next_position = self.get_next_center_position()
    
        # Check if there is an available position
        if next_position is None or next_position >= self.max_speakers:
            raise IndexError("No available positions to add a new centroid.")
        self.centers[next_position] = embedding
        return next_position

    def identify_single_speaker(self, embedding: torch.Tensor) -> int:
        """Identify the identity of a single speaker

        Parameters:
        embedding: torch.Tensor
            The embedding of the speaker

        Returns:
        global_speaker: int
            The global speaker index corresponding to the speaker
        """

        # Initialize centroids if they don't exist yet
        if self.centers is None:
            self.init_centers(embedding.shape[0])
            global_speaker = self.add_center(embedding)
            return global_speaker
        
        # Calculate distance between the embedding and all existing centroids
        distances = cdist(embedding.reshape(1, -1), self.centers, metric=self.metric).flatten()

        # If we've reached the maximum number of centroids, we can only assign the data to the nearest centroid
        if self.get_next_center_position() is None:
            # Directly find and return the nearest centroid
            global_speaker = np.argmin(distances)
            # Do not update the centroid
            return global_speaker
        
        # Check if any existing centroid is within the threshold distance
        valid_map = np.where(distances < self.delta_new)[0]

        # If there are centroids within threshold, choose the closest one
        if valid_map.size > 0:
            # Choose the closest one
            global_speaker = valid_map[np.argmin(distances[valid_map])]
            # Update the centroid
            self.update(global_speaker, embedding)
            return global_speaker
        
        # If no centroid is close enough, add a new one
        global_speaker = self.add_center(embedding)
        self.counts[global_speaker] += 1

        return global_speaker
