from rx import operators as ops
from diart.sources import MicrophoneAudioSource
from pyannote.audio.pipelines.speaker_verification import PretrainedSpeakerEmbedding
from km_clustering import IncrementalSpeakerClustering
import diart.operators as dops
import whisper
import torch
import numpy as np
import sys

class OnlineDiarizeASR:
    def __init__(self, **kwargs):
        '''
        Parameters:

        [Audio Processing Parameters]
        - step (float): The step length of speech segmentation (in seconds), default is 0.5.
        - duration (float): The duration of each speech segment (in seconds), default is 0.5.
        - no_speech_prob_threshold (float): The probability threshold for a segment to be considered as non-speech, default is 0.6.

        [Whisper Parameters]
        - whisper_model (str): The name of the speech recognition model, default is 'base'.
        - language (str): The language for speech recognition, default is None.
        - initial_prompt (str): The initial prompt for speech recognition, default is None.

        [Diarization Parameters]
        - n_speakers (int): The maximum number of speakers expected in the input, default is 20.
        - delta_new (float): The maximum cosine distance between two speech segments, default is 0.8.
        '''

        # Audio processing parameters
        self.sample_rate = 16000
        self.step = kwargs.get('step', 0.5)
        self.duration = kwargs.get('duration', 0.5)
        self.no_speech_prob_threshold = kwargs.get('no_speech_prob_threshold', 0.6)
        
        # Whisper parameters
        self.whisper_model = kwargs.get('whisper_model', 'base')
        self.language = kwargs.get('language', None)
        self.initial_prompt = kwargs.get('initial_prompt', None)

        # Diarization parameters
        self.n_speakers = kwargs.get('n_speakers', 20)
        self.delta_new = kwargs.get('delta_new', 0.8)

        # Models and microphone
        self.mic = MicrophoneAudioSource(self.sample_rate)
        self.asr_model = whisper.load_model(self.whisper_model)
        self.emb_model = PretrainedSpeakerEmbedding(
            "speechbrain/spkrec-ecapa-voxceleb",
            device=torch.device("cuda" if torch.cuda.is_available() else "cpu")
        )
        self.cluster_model = IncrementalSpeakerClustering(
            delta_new=self.delta_new,
            max_speakers=self.n_speakers
        )
        
        # Internal state parameters
        self.audio_buffer = np.array([], dtype=np.float32)
        self.transcription_results = []
        self.buffer_popped = False
        self.prev_segments_num = 1
        self.current_speaker = ''

    def update_audio_and_transcribe(self, audio_chunk: np.ndarray):
        # Add new audio chunk to the audio buffer
        self.audio_buffer = np.concatenate((self.audio_buffer, audio_chunk))

        # Recognize the entire audio buffer
        result = self.asr_model.transcribe(
            self.audio_buffer,
            word_timestamps=True,
            initial_prompt=self.initial_prompt, 
            language=self.language
        )

        if len(result['segments']) > 2:
            # Calculate the length of the audio buffer to pop based on segment timestamps
            samples_to_pop = int(result['segments'][0]['end'] * self.sample_rate)
            reshaped_samples = torch.from_numpy(self.audio_buffer[:samples_to_pop][None,None])
            embedding = self.emb_model(reshaped_samples)
            self.current_speaker = self.cluster(embedding.squeeze())
            # Pop the audio buffer
            self.audio_buffer = self.audio_buffer[samples_to_pop:]
            self.buffer_popped = True
        else:
            self.buffer_popped = False
        return result

    def is_speech(self, result: dict):
        if len(result['segments']) == 0:
            return False
        return result['segments'][-1]['no_speech_prob'] < self.no_speech_prob_threshold

    def print_result(self, result: dict):
        # Clear the previous result from bottom to top
        move_up_and_clear = '\r\033[K' + '\033[A\033[K'*(self.prev_segments_num-1)
        sys.stdout.write(move_up_and_clear)

        output_list = [f"{seg['text']}" for seg in result['segments']]
        self.prev_segments_num = len(output_list)

        if self.buffer_popped:
            # Print with color coding for different speakers
            print(f"\033[{self.current_speaker+31}m{output_list[0]}\033[0m")
            if len(output_list) > 1:
                joined_output = "\n".join(output_list[1:])
                self.prev_segments_num -= 1
                print(f'\r{joined_output}', end="", flush=True)
        else:
            # Overwrite the current line
            joined_output = "\n".join(output_list)
            print(f'\r{joined_output}', end="", flush=True)

    def is_silence(self, audio_chunk: np.ndarray):
        return np.all(audio_chunk == 0)

    def cluster(self, embeddings: torch.Tensor):
        label = self.cluster_model.identify_single_speaker(embeddings)
        return label
    
    def __call__(self):
        # Subscribe to the microphone audio stream
        self.mic.stream.pipe(
            dops.rearrange_audio_stream(
                duration=self.duration,
                step=self.step,
                sample_rate=self.sample_rate
            ),
            ops.map(lambda x: np.concatenate(x)),
            ops.filter(lambda audio_data: not self.is_silence(audio_data)),
            ops.map(self.update_audio_and_transcribe),
            ops.filter(self.is_speech),
            ops.do_action(self.print_result)
        ).subscribe(
            on_next=lambda _: None,
            on_error=lambda e: print(e)
        )

        print("[Starting]")
        self.mic.read()

if __name__ == '__main__':
    # Configuration parameters
    config = {
        'n_speakers': 10, # max_speakers
        'no_speech_prob_threshold': 0.5, 
        'delta_new': 0.8, # maximum cosine distance between two speech segments
        'whisper_model': 'base',
    }

    online_diarize_asr = OnlineDiarizeASR(**config)
    online_diarize_asr()
